from typing import Optional

from mavtel_models.mavlink.metrics import *


class MetricsMixin:
    """
    This is an autogenerated class!
    """

    async def actuator_control_target(self) -> Optional[ActuatorControlTarget]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(ActuatorControlTarget)

    async def actuator_output_status(self) -> Optional[ActuatorOutputStatus]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(ActuatorOutputStatus)

    async def attitude_angular_velocity_body(self) -> Optional[AttitudeAngularVelocityBody]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(AttitudeAngularVelocityBody)

    async def attitude_euler(self) -> Optional[AttitudeEuler]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(AttitudeEuler)

    async def attitude_quaternion(self) -> Optional[AttitudeQuaternion]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(AttitudeQuaternion)

    async def armed(self) -> Optional[Armed]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Armed)

    async def battery(self) -> Optional[Battery]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Battery)

    async def camera_attitude_euler(self) -> Optional[CameraAttitudeEuler]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(CameraAttitudeEuler)

    async def camera_attitude_quaternion(self) -> Optional[CameraAttitudeQuaternion]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(CameraAttitudeQuaternion)

    async def distance_sensor(self) -> Optional[DistanceSensor]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(DistanceSensor)

    async def fixedwing_metrics(self) -> Optional[FixedwingMetrics]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(FixedwingMetrics)

    async def flight_mode(self) -> Optional[FlightMode]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(FlightMode)

    async def gps_info(self) -> Optional[GpsInfo]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(GpsInfo)

    async def ground_truth(self) -> Optional[GroundTruth]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(GroundTruth)

    async def heading(self) -> Optional[Heading]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Heading)

    async def health(self) -> Optional[Health]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Health)

    async def health_all_ok(self) -> Optional[HealthAllOk]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(HealthAllOk)

    async def home(self) -> Optional[Home]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Home)

    async def imu(self) -> Optional[Imu]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Imu)

    async def in_air(self) -> Optional[InAir]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(InAir)

    async def landed_state(self) -> Optional[LandedState]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(LandedState)

    async def odometry(self) -> Optional[Odometry]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Odometry)

    async def position(self) -> Optional[Position]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(Position)

    async def position_velocity_ned(self) -> Optional[PositionVelocityNed]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(PositionVelocityNed)

    async def raw_gps(self) -> Optional[RawGps]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(RawGps)

    async def raw_imu(self) -> Optional[RawImu]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(RawImu)

    async def rc_status(self) -> Optional[RcStatus]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(RcStatus)

    async def unix_epoch_time(self) -> Optional[UnixEpochTime]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(UnixEpochTime)

    async def velocity_ned(self) -> Optional[VelocityNed]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(VelocityNed)

    async def vtol_state(self) -> Optional[VtolState]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class(VtolState)



from typing import Any, Dict, Union

import orjson

from mavtel_models.websockets.message_type import WSMessageType
from mavtel_models.websockets.messages.base_message import BaseWSMessage
from mavtel_models.websockets.messages.heartbeat import HeartbeatWSMessage
from mavtel_models.websockets.messages.metric_update import MetricsUpdateWSMessage
from mavtel_models.websockets.messages.subscribe import SubscribeWSMessage


def get_ws_message_classes() -> Dict[WSMessageType, BaseWSMessage.__class__]:
    return {
        WSMessageType.HEARTBEAT: HeartbeatWSMessage,
        WSMessageType.METRIC_UPDATE: MetricsUpdateWSMessage,
        WSMessageType.SUBSCRIBE: SubscribeWSMessage,
    }


def parse_ws_message(data: Union[str, Dict[str, Any]]) -> BaseWSMessage:
    data: Dict[str, Any] = orjson.loads(data) if isinstance(data, str) else data

    message_type = WSMessageType(data['type'])

    return get_ws_message_classes()[message_type](**data)



import httpx

from mavtel_models import SessionState, CollectorState, StorageState


class ServicesMixin:
    """
    This is an autogenerated class!
    """

    async def session(self) -> SessionState:
        # noinspection PyUnresolvedReferences
        url = self._api_url + '/services/session'
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(url)
            return SessionState(**response.json())

    async def collector(self) -> CollectorState:
        # noinspection PyUnresolvedReferences
        url = self._api_url + '/services/collector'
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(url)
            return CollectorState(**response.json())

    async def storage(self) -> StorageState:
        # noinspection PyUnresolvedReferences
        url = self._api_url + '/services/storage'
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(url)
            return StorageState(**response.json())



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.location import Location


class Position(Location, MAVLinkMetricBaseModel):
    relative_altitude_m: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Position'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.landed_state import LandedState as _LandedState


class LandedState(MAVLinkMetricBaseModel):
    landed_state: _LandedState

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(landed_state=_LandedState(mavlink_object.value))

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.LandedState'

    def as_flat_dict(self):
        return dict(
            landed_state=self.landed_state.value
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class FixedwingMetrics(MAVLinkMetricBaseModel):
    airspeed_m_s: float
    throttle_percentage: float
    climb_rate_m_s: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.FixedwingMetrics'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class Heading(MAVLinkMetricBaseModel):
    heading_deg: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Heading'



from typing import List

from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel, FLAT_STRUCT_DELIMITER


class ActuatorControlTarget(MAVLinkMetricBaseModel):
    group: int
    controls: List[float]

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(group=mavlink_object.group, controls=list(mavlink_object.controls))

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.ActuatorControlTarget'

    def as_flat_dict(self):
        return dict(
            active=self.group,
            **{
                FLAT_STRUCT_DELIMITER.join(['controls', str(i)]): self.controls[i]
                for i in range(len(self.controls))
            }
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class Battery(MAVLinkMetricBaseModel):
    id: int
    voltage_v: float
    remaining_percent: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Battery'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class DistanceSensor(MAVLinkMetricBaseModel):
    minimum_distance_m: float
    maximum_distance_m: float
    current_distance_m: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.DistanceSensor'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.vtol_state import VtolState as _VtolState


class VtolState(MAVLinkMetricBaseModel):
    vtol_state: _VtolState

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(vtol_state=_VtolState(mavlink_object.value))

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.VtolState'

    def as_flat_dict(self):
        return dict(
            vtol_state=self.vtol_state.value
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class InAir(MAVLinkMetricBaseModel):
    in_air: bool

    @classmethod
    def from_mavlink(cls, mavlink_object: int):
        return cls(in_air=mavlink_object)

    def get_mavlink_data_class_name(self) -> str:
        return 'bool'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel, FLAT_STRUCT_DELIMITER
from mavtel_models.mavlink.primitives.acceleration_frd import AccelerationFrd
from mavtel_models.mavlink.primitives.angular_velocity_frd import AngularVelocityFrd
from mavtel_models.mavlink.primitives.magnetic_field_frd import MagneticFieldFrd


class Imu(MAVLinkMetricBaseModel):
    acceleration_frd: AccelerationFrd
    angular_velocity_frd: AngularVelocityFrd
    magnetic_field_frd: MagneticFieldFrd
    temperature_degc: float
    timestamp_us: int

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(
            acceleration_frd=mavlink_object.acceleration_frd.__dict__,
            angular_velocity_frd=mavlink_object.angular_velocity_frd.__dict__,
            magnetic_field_frd=mavlink_object.magnetic_field_frd.__dict__,
            temperature_degc=mavlink_object.temperature_degc,
            timestamp_us=mavlink_object.timestamp_us,
        )

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Imu'

    def as_flat_dict(self):
        return dict(
            **{
                FLAT_STRUCT_DELIMITER.join(['acceleration_frd', name]): value
                for name, value in self.acceleration_frd.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['angular_velocity_frd', name]): value
                for name, value in self.angular_velocity_frd.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['magnetic_field_frd', name]): value
                for name, value in self.magnetic_field_frd.dict().items()
            },
            temperature_degc=self.temperature_degc,
            timestamp_us=self.timestamp_us,
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class HealthAllOk(MAVLinkMetricBaseModel):
    health_all_ok: bool

    @classmethod
    def from_mavlink(cls, mavlink_object: bool):
        return cls(health_all_ok=mavlink_object)

    def get_mavlink_data_class_name(self) -> str:
        return 'bool'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel, FLAT_STRUCT_DELIMITER
from mavtel_models.mavlink.primitives.angular_velocity_body import AngularVelocityBody
from mavtel_models.mavlink.primitives.covariance import Covariance
from mavtel_models.mavlink.primitives.mav_frame import MavFrame
from mavtel_models.mavlink.primitives.position_body import PositionBody
from mavtel_models.mavlink.primitives.quaternion import Quaternion
from mavtel_models.mavlink.primitives.velocity_body import VelocityBody


class Odometry(MAVLinkMetricBaseModel):
    time_usec: int
    frame_id: MavFrame
    child_frame_id: MavFrame
    position_body: PositionBody
    q: Quaternion
    velocity_body: VelocityBody
    angular_velocity_body: AngularVelocityBody
    pose_covariance: Covariance
    velocity_covariance: Covariance

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(
            time_usec=mavlink_object.time_usec,
            frame_id=MavFrame(mavlink_object.frame_id.value),
            child_frame_id=MavFrame(mavlink_object.child_frame_id.value),
            position_body=mavlink_object.position_body.__dict__,
            q=mavlink_object.q.__dict__,
            velocity_body=mavlink_object.velocity_body.__dict__,
            angular_velocity_body=mavlink_object.angular_velocity_body.__dict__,
            pose_covariance=Covariance(
                covariance_matrix=list(mavlink_object.pose_covariance.covariance_matrix)
            ),
            velocity_covariance=Covariance(
                covariance_matrix=list(mavlink_object.velocity_covariance.covariance_matrix)
            ),
        )

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Odometry'

    def as_flat_dict(self):
        return dict(
            **{
                FLAT_STRUCT_DELIMITER.join(['position_body', name]): value
                for name, value in self.position_body.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['q', name]): value
                for name, value in self.q.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['velocity_body', name]): value
                for name, value in self.velocity_body.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['angular_velocity_body', name]): value
                for name, value in self.velocity_body.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['pose_covariance', str(i)]): self.pose_covariance.covariance_matrix[i]
                for i in range(len(self.pose_covariance.covariance_matrix))
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['velocity_covariance', str(i)]): self.velocity_covariance.covariance_matrix[
                    i]
                for i in range(len(self.velocity_covariance.covariance_matrix))
            },
            time_usec=self.time_usec,
            frame_id=self.frame_id.value,
            child_frame_id=self.child_frame_id.value,
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class VelocityNed(MAVLinkMetricBaseModel):
    north_m_s: float
    east_m_s: float
    down_m_s: float

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.VelocityNed'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel, FLAT_STRUCT_DELIMITER
from mavtel_models.mavlink.primitives.position_ned import PositionNed
from mavtel_models.mavlink.primitives.velocity_ned import VelocityNed


class PositionVelocityNed(MAVLinkMetricBaseModel):
    position: PositionNed
    velocity: VelocityNed

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(
            position=mavlink_object.position.__dict__,
            velocity=mavlink_object.velocity.__dict__,
        )

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.PositionVelocityNed'

    def as_flat_dict(self):
        return dict(
            **{
                FLAT_STRUCT_DELIMITER.join(['position', name]): value
                for name, value in self.position.dict().items()
            },
            **{
                FLAT_STRUCT_DELIMITER.join(['velocity', name]): value
                for name, value in self.velocity.dict().items()
            },
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class Health(MAVLinkMetricBaseModel):
    is_gyrometer_calibration_ok: bool
    is_accelerometer_calibration_ok: bool
    is_magnetometer_calibration_ok: bool
    is_local_position_ok: bool
    is_global_position_ok: bool
    is_home_position_ok: bool
    is_armable: bool

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Health'



import os

BASE_DIR = os.path.dirname(os.path.realpath(__file__))


def generate_client_metrics_mixin():
    print('Generating "mavtel_client.client.mixins.metrics_mixin"...')

    from mavtel_models.supported_metrics import get_supported_metrics

    supported_metrics = get_supported_metrics()

    output = '''from typing import Optional
    
from mavtel_models.mavlink.metrics import *
    
    
class MetricsMixin:
    """
    This is an autogenerated class!
    """
    '''

    for metric_name, metric_class in supported_metrics.items():
        metric_class_name = metric_class.__name__

        output += f"""
    async def {metric_name}(self) -> Optional[{metric_class_name}]:
        # noinspection PyUnresolvedReferences
        return await self.get_metric_by_class({metric_class_name})
        """

    with open(os.path.join(BASE_DIR, 'client', 'mixins', 'metrics_mixin.py'), mode='wt') as fs:
        fs.write(output)


def generate_client_mixins_services_mixin():
    print('Generating "mavtel_client.client.mixins.services_mixin"...')

    services = dict(
        session='SessionState',
        collector='CollectorState',
        storage='StorageState',
    )

    output = f'''import httpx

from mavtel_models import {", ".join(services.values())}
    
    
class ServicesMixin:
    """
    This is an autogenerated class!
    """
    '''

    for service_name, service_state_class in services.items():
        output += f"""
    async def {service_name}(self) -> {service_state_class}:
        # noinspection PyUnresolvedReferences
        url = self._api_url + '/services/{service_name}'
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(url)
            return {service_state_class}(**response.json())
        """

    with open(os.path.join(BASE_DIR, 'client', 'mixins', 'services_mixin.py'), mode='wt') as fs:
        fs.write(output)


if __name__ == "__main__":
    generate_client_metrics_mixin()
    generate_client_mixins_services_mixin()



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.flight_mode import FlightMode as _FlightMode


class FlightMode(MAVLinkMetricBaseModel):
    flight_mode: _FlightMode

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(flight_mode=_FlightMode(mavlink_object.value))

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.FlightMode'

    def as_flat_dict(self):
        return dict(
            flight_mode=self.flight_mode.value
        )



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.angular_velocity_body import AngularVelocityBody


class AttitudeAngularVelocityBody(AngularVelocityBody, MAVLinkMetricBaseModel):
    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.AngularVelocityBody'



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.quaternion import Quaternion


class AttitudeQuaternion(Quaternion, MAVLinkMetricBaseModel):
    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.Quaternion'



from typing import Any, Optional

from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class RcStatus(MAVLinkMetricBaseModel):
    was_available_once: bool
    is_available: bool
    signal_strength_percent: Optional[float]

    def __init__(self, **data: Any):
        super().__init__(**data)

        if self.signal_strength_percent is None:
            self.signal_strength_percent = float('nan')

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.RcStatus'



from typing import List

from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel, FLAT_STRUCT_DELIMITER


class ActuatorOutputStatus(MAVLinkMetricBaseModel):
    active: int
    actuator: List[float]

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(active=mavlink_object.active, actuator=list(mavlink_object.actuator))

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.ActuatorOutputStatus'

    def as_flat_dict(self):
        return dict(
            active=self.active,
            **{
                FLAT_STRUCT_DELIMITER.join(['actuator', str(i)]): self.actuator[i]
                for i in range(len(self.actuator))
            }
        )



from .api_info import APIInfo
from .app_ready_response import AppReadyResponse
from .collector_state import CollectorState
from .session_state import SessionState
from .storage_state import StorageState



from enum import Enum


class FlightMode(Enum):
    UNKNOWN = 0
    READY = 1
    TAKEOFF = 2
    HOLD = 3
    MISSION = 4
    RETURN_TO_LAUNCH = 5
    LAND = 6
    OFFBOARD = 7
    FOLLOW_ME = 8
    MANUAL = 9
    ALTCTL = 10
    POSCTL = 11
    ACRO = 12
    STABILIZED = 13
    RATTITUDE = 14



import logging
from typing import Optional, Set, Union
from urllib.parse import urljoin, urlparse

import httpx
import orjson
import websockets
from websockets.exceptions import ConnectionClosed

from mavtel_models import MAVLinkMetricBaseModel, SubscribeWSMessage, MetricsUpdateWSMessage
from mavtel_models import get_supported_metrics, parse_ws_message

log = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 50


class MAVTelClientBase:
    def __init__(self, url: str):
        self._url: str = url
        self._api_url = urljoin(self._url, 'api/v1')
        self._supported_metrics = get_supported_metrics()

    @property
    def url(self) -> str:
        return self._url

    def get_metric_url(self, metric: Union[MAVLinkMetricBaseModel.__class__, str]) -> str:
        metric_name = metric if isinstance(metric, str) else metric.get_metric_name()
        return f'{self._api_url}/metrics/{metric_name}'

    def get_metric_updates_ws_url(self) -> str:
        parsed_url = urlparse(self._url)
        ws_base_url = parsed_url._replace(scheme='ws').geturl()
        return urljoin(ws_base_url, 'api/v1/streams/metrics')

    async def healthcheck(self) -> bool:
        url = urljoin(self._url, 'health/check')
        async with httpx.AsyncClient() as client:
            try:
                response: httpx.Response = await client.get(url)
                return response.status_code == 200 and response.text == 'OK'
            except httpx.ConnectError:
                return False

    async def get_metric_by_class(
            self,
            metric_class: MAVLinkMetricBaseModel.__class__
    ) -> Optional[MAVLinkMetricBaseModel]:
        url = self.get_metric_url(metric_class)
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(url, timeout=DEFAULT_TIMEOUT)
            data = orjson.loads(response.text)
            if data is None:
                return None
            return metric_class(**data)

    async def metric_updates(self, metric_names: Set[str] = None):
        metric_names = metric_names or set(self._supported_metrics.keys())
        if len(metric_names.difference(self._supported_metrics)) > 0:
            raise KeyError(f'Metrics {", ".join(metric_names.difference(self._supported_metrics))} are not supported!')

        async with websockets.connect(self.get_metric_updates_ws_url()) as websocket:
            subscribe_msg = SubscribeWSMessage(
                metrics={name: None for name in metric_names}
            ).json()
            await websocket.send(subscribe_msg)

            while True:
                try:
                    raw_message = await websocket.recv()
                    # noinspection PyTypeChecker
                    message: MetricsUpdateWSMessage = parse_ws_message(raw_message)

                    if not isinstance(message, MetricsUpdateWSMessage):
                        continue

                    metric_class: MAVLinkMetricBaseModel.__class__ = self._supported_metrics[message.name]

                    yield metric_class(**message.data)
                except orjson.JSONDecodeError as json_error:
                    log.error(f"Error decoding metrics message: {json_error}. Message: '{raw_message}'")
                except ConnectionClosed:
                    log.error(f'Websocket connection closed')
                    break



from typing import Dict

from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.metrics import *


def get_supported_metrics() -> Dict[str, MAVLinkMetricBaseModel.__class__]:
    return dict(
        actuator_control_target=ActuatorControlTarget,
        actuator_output_status=ActuatorOutputStatus,
        attitude_angular_velocity_body=AttitudeAngularVelocityBody,
        attitude_euler=AttitudeEuler,
        attitude_quaternion=AttitudeQuaternion,
        armed=Armed,
        battery=Battery,
        camera_attitude_euler=CameraAttitudeEuler,
        camera_attitude_quaternion=CameraAttitudeQuaternion,
        distance_sensor=DistanceSensor,
        fixedwing_metrics=FixedwingMetrics,
        flight_mode=FlightMode,
        gps_info=GpsInfo,
        ground_truth=GroundTruth,
        heading=Heading,
        health=Health,
        health_all_ok=HealthAllOk,
        home=Home,
        imu=Imu,
        in_air=InAir,
        landed_state=LandedState,
        odometry=Odometry,
        position=Position,
        position_velocity_ned=PositionVelocityNed,
        raw_gps=RawGps,
        raw_imu=RawImu,
        rc_status=RcStatus,
        unix_epoch_time=UnixEpochTime,
        velocity_ned=VelocityNed,
        vtol_state=VtolState,
    )



from typing import Set

from inflection import underscore
from pydantic import BaseModel

FLAT_STRUCT_DELIMITER = '___'


class MAVLinkMetricBaseModel(BaseModel):
    @classmethod
    def from_mavlink(cls, mavlink_object):
        return cls(**mavlink_object.__dict__)

    @classmethod
    def get_metric_name(cls) -> str:
        return underscore(cls.__name__)

    @classmethod
    def get_list_fields_names(cls) -> Set[str]:
        return set()

    def get_mavlink_data_class_name(self) -> str:
        raise NotImplemented

    def as_flat_dict(self):
        return self.dict()



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class RawGps(MAVLinkMetricBaseModel):
    timestamp_us: int
    latitude_deg: float
    longitude_deg: float
    absolute_altitude_m: float
    hdop: float
    vdop: float
    velocity_m_s: float
    cog_deg: float
    altitude_ellipsoid_m: float
    horizontal_uncertainty_m: float
    vertical_uncertainty_m: float
    velocity_uncertainty_m_s: float
    heading_uncertainty_deg: float
    yaw_deg: float

    def get_mavlink_data_class_name(self) -> object:
        return 'mavsdk.telemetry.RawGps'



from typing import Dict

from pydantic import BaseModel

from mavtel_models.typing import BackendConfigType


class StorageBackendState(BaseModel):
    is_started: bool
    is_ready: bool
    backend_type: str
    config: BackendConfigType


class StorageState(BaseModel):
    is_ready: bool
    is_started: bool
    backends: Dict[str, StorageBackendState]



"""
MAVTel package setup
    Keywords: https://setuptools.pypa.io/en/latest/references/keywords.html
    A good example of setup.py can be found here: https://github.com/pypa/sampleproject/blob/main/setup.py
    Packaging HOWTO: https://packaging.python.org/en/latest/tutorials/packaging-projects/
"""
import os
import subprocess

from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))

try:
    with open(os.path.join(ROOT_DIR, '.VERSION')) as f:
        version = f.readlines()[0]
except FileNotFoundError:
    version = subprocess.check_output('git describe --tags --abbrev=0', cwd=ROOT_DIR, shell=True)

with open(os.path.join(ROOT_DIR, 'mavtel_client', 'README.md'), encoding='utf-8') as fs:
    long_description = fs.read()

test_dependencies = [
    'aiounittest>=1.4.2',
    'coverage>=6.4',
    'pytest>=7.1',
]

extras = {
    'test': test_dependencies,
}

setup(
    name='mavtel-client',
    version=version,
    description='Client for MAVTel',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    author='Mykhailo Ziatin',
    author_email='mikhail.zyatin@gmail.com',
    packages=find_packages(
        where='.',
        include=['mavtel_models**', 'mavtel_client**'],
    ),
    package_dir={'': '.'},
    python_requires='>=3.7, <4',
    install_requires=[
        'httpx>=0.23.0',
        'inflection>=0.5.1',
        'orjson>=3.7.7',
        'pydantic>=1.9.1',
        'websockets>=10.3',
    ],
    setup_requires=[],
    license_files=('LICENSE',),
    tests_require=test_dependencies,
    extras_require=extras,
    data_files=[('version', ['.VERSION', '.BUILD']), ('docs', ['mavtel_client/README.md'])],
    url='https://gitlab.com/Zyatin/mavtel/mavtel_client',
    project_urls={
        'Bug Reports': 'https://gitlab.com/Zyatin/mavtel/-/issues',
        'Source': 'https://gitlab.com/Zyatin/mavtel',
        'Homepage': 'https://gitlab.com/Zyatin/mavtel/mavtel_client',
    },
)



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel
from mavtel_models.mavlink.primitives.fix_type import FixType


class GpsInfo(MAVLinkMetricBaseModel):
    num_satellites: int
    fix_type: FixType

    @classmethod
    def from_mavlink(cls, mavlink_object: object):
        # noinspection PyUnresolvedReferences
        return cls(
            num_satellites=mavlink_object.num_satellites,
            fix_type=FixType(mavlink_object.fix_type.value)
        )

    def get_mavlink_data_class_name(self) -> str:
        return 'mavsdk.telemetry.GpsInfo'

    def as_flat_dict(self):
        return dict(
            num_satellites=self.num_satellites,
            fix_type=self.fix_type.value,
        )



from .actuator_control_target import ActuatorControlTarget
from .actuator_output_status import ActuatorOutputStatus
from .armed import Armed
from .attitude_angular_velocity_body import AttitudeAngularVelocityBody
from .attitude_euler import AttitudeEuler
from .attitude_quaternion import AttitudeQuaternion
from .battery import Battery
from .camera_attitude_euler import CameraAttitudeEuler
from .camera_attitude_quaternion import CameraAttitudeQuaternion
from .distance_sensor import DistanceSensor
from .fixedwing_metrics import FixedwingMetrics
from .flight_mode import FlightMode
from .gps_info import GpsInfo
from .ground_truth import GroundTruth
from .heading import Heading
from .health import Health
from .health_all_ok import HealthAllOk
from .home import Home
from .imu import Imu
from .in_air import InAir
from .landed_state import LandedState
from .odometry import Odometry
from .position import Position
from .position_velocity_ned import PositionVelocityNed
from .raw_gps import RawGps
from .raw_imu import RawImu
from .rc_status import RcStatus
from .unix_epoch_time import UnixEpochTime
from .velocity_ned import VelocityNed
from .vtol_state import VtolState



from typing import Dict, Optional

from mavtel_models.websockets.message_type import WSMessageType
from mavtel_models.websockets.messages.base_message import BaseWSMessage


class SubscribeWSMessage(BaseWSMessage):
    type = WSMessageType.SUBSCRIBE
    metrics: Dict[str, Optional[float]]



from mavtel_client.client.base import MAVTelClientBase
from mavtel_client.client.mixins.metrics_mixin import MetricsMixin
from mavtel_client.client.mixins.services_mixin import ServicesMixin


class MAVTelClient(MAVTelClientBase, MetricsMixin, ServicesMixin):
    pass



from mavtel_models.mavlink.base_metric_model import MAVLinkMetricBaseModel


class UnixEpochTime(MAVLinkMetricBaseModel):
    unix_epoch_time: int

    @classmethod
    def from_mavlink(cls, mavlink_object: int):
        return cls(unix_epoch_time=mavlink_object)

    def get_mavlink_data_class_name(self) -> str:
        return 'int'
